/**
 * This program instruments DSP filters for testing speed and accuracy.
 * Threads on core0:
 * --graphics thread draws the bode plots
 * --serial thread gets frequency sweep range
 * --blinky
 * ISR on core1:
 * -- runs at 40 KHz to generate a sweep sinewave by DDS, then filter it.
 * The filter output is cross correlated with the sine (and cosine) to get the
 * effect on phase and magnitude.
 * -- optional ADC input and SPI DAC output of realtime audio
 * -- Without SPI DAC output ISR duration is 2.5 uSec, with the 2-pole IIR taking 1.1 uSec of that.
 * -- With SPI DAC output the duration is 8 uSec.
 * 
 * FIR version
 * Commands:
 * help
 * 
 * window type (rect/hamming/hann/blackman)
 * lpw frequency length
 * --------------
 * color int repesenting one of 7 primaries
 * freq Flo Fhi Finc 
 * downsample yes/no
 * erase
 * dac
 * realtime
 * 
 * VGA by Hunter Adams (vha3@cornell.edu)
 * converted to 320x240 with 256 colors by Bruce; bruce.land@cornell.edu
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 * 
 *  - GPIO 08 ---> 330 ohm resistor ---> VGA Blue lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 09 ---> 220 ohm resistor ---> VGA Blue hi_bit |   and to VGA blue
 * 
 *  - GPIO 10 ---> 1000 ohm resistor ---> VGA Green lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 11 ---> 680 ohm resistor ---> VGA Green mid_bit |   and to VGA Green
 *  - GPIO 12 ---> 330 ohm resistor ---> VGA Green hi_bit  |   
 * 
 *  - GPIO 13 ---> 1000 ohm resistor ---> VGA Red lo-bit |__ three wired to 100 ohm to ground
 *  - GPIO 14 ---> 680 ohm resistor ---> VGA Red mid_bit |   and to VGA red
 *  - GPIO 15 ---> 330 ohm resistor ---> VGA Red hi_bit  |   
 * 
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0 to 3 on PIO instance 0
 *  - DMA channels 0, 1, 2, 3 data send to two PIO
 *  - 76.8 kBytes of RAM (for pixel color data)
 * color encoding: bits 7:5 red; 4:2 green; 1:0 blue
 *
 * Protothreads v1.1.1
 * graphics dsp plot thread
 * serial thread to set the filter parameters///
 * the usual blinky thread for a hearbeat
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga256_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "hardware/adc.h"

//===========================================
// === other hardware
//===========================================
#include "hardware/gpio.h"
#include "hardware/timer.h"
#include "hardware/spi.h"
#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "hardware/gpio.h"
#include "hardware/structs/iobank0.h"
#include "hardware/regs/rosc.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
#include "math.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"
// interactive flag
int new_value = false ;

// ========================================
// === dds setup 
// =======================================
//
// == bit fixed point 2.14 format ===========
// resolution 2^-14 =  6.1035e-5
// dynamic range is +1.9999/-2.0
typedef signed short s1x14 ;
#define muls1x14(a,b) ((s1x14)((((int32_t)(a))*((int32_t)(b)))>>14)) //multiply two fixed s1x14
#define float_to_s1x14(a) ((s1x14)((a)*16384.0)) // 2^14
#define s1x14_to_float(a) ((float)(a)/16384.0)
#define absfix14(a) abs(a)

// =======================================
// 1/Fs in microseconds for ISR
volatile int alarm_period = 25 ;

// DDS variables 
unsigned int main_inc ;
unsigned int main_accum ;
s1x14 sine_table_out, cos_table_out ;
s1x14  filter_out, down_sampled_signal ;
s1x14 sine_table[256], cos_table[256] ;
// rates and freqs
float Fs=40000, Fout, F_low=10, F_high=20000, F_inc=1.1;
float lin_Fout, lin_F_low=100, lin_F_high=10000, lin_F_inc=50;

// input command string
char str_in[30] ;
// filter design inputs
float F0, F1, F2,  w0, w1, w2, M ;
// parameters for  filter design
s1x14  b[128] ;
float  b_float[128] ;
int filter_length ;
//r--recanglular; h--hamming; b-blackman
char window_type ;
float wind_val ;


// interatie control of filter
int decimate, filter_type ;
// modulo countr for downsampling in CIC
 int downsample_clock ;
 //
 char plot_color = GREEN ;


// cross correlation
// phase/amp variables
volatile int I_signal, Q_signal ;
// averaging the signal
int avg_size = 10000;
volatile int avg_count;
// accumulators
volatile int32_t I_avg, Q_avg;
// and the final averages
float I_out, Q_out, mag, phase, I ;//
// control realtime audio
int audio_flag, dac_flag ;
int adc_in ;
s1x14 filter_input ;

// ========================================
// === spi setup 
// =======================================
//SPI configurations
#define PIN_CS   5
#define PIN_SCK  6
#define PIN_MOSI 7
#define SPI_PORT spi0

// constant to tell SPI DAC what to do
// prepend to each 12-bit sample
#define DAC_config_chan_A 0b0011000000000000
// B-channel, 1x, active
#define DAC_config_chan_B 0b1011000000000000
uint16_t DAC_data ; 
//
// ==================================================
// === graphics plot bode -- RUNNING on core 0
// ==================================================
        
int start_time, end_time ;
char video_string[32] ;
// graph labels
short decade_name[4] = {10, 100, 1000, 10000};
// short mag_name[5] = {0, -1, -2, -3 , -4, -5};
float mag_name[5] = {1, .1, .01, .001, .0001};

static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // background
    fillRect(0, 0, 319, 239, BLACK); // 

    // Draw some filled rectangles
    fillRect(0, 0, 76, 10, BLUE); // blue box
    fillRect(100, 0, 150, 10, GREEN); // red box
    //fillRect(200, 0, 76, 10, GREEN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(10, 1) ;
    setTextSize(1) ;
    writeString("ECE 4760") ;

    setTextColor(BLACK) ;
    setCursor(102, 1) ;
    setTextSize(1) ;
    writeString("DSP Bode Plots s1x14") ;

    // graphics constants in pixels
    // for plots
    #define decade_width 75 // pixels
    #define decade_start 20
    #define zero_phase 170 // y-cood
    #define gain_phase 0.2 // pixels/degree
    #define zero_mag 25 // y cood 0 db
    #define gain_mag 20 // pixels/decade
    
    while(true) {

        // plot new curve
        if(new_value == true){  
          
          // now set up the DSP
          // force realtime off
          if (audio_flag){
            audio_flag = 0 ;
            printf("*** Realtime mode forced OFF ***\n\r");
          }
            Fout = F_low ;
            //printf("freq,   mag,  phase\n\r") ;
            //
            
            //
            static int i ;
            while(Fout <= F_high){
              // set dds
                main_inc = Fout * pow(2,32 )/ Fs ;
                // the measurement
                I_avg = 0 ;
                Q_avg = 0 ;
                avg_count = 0;
                // wait for # of samples
                while (avg_count<avg_size) {}; 
                // and get the average
                I_out = s1x14_to_float(I_avg / avg_size );
                Q_out = s1x14_to_float(Q_avg / avg_size );

                mag = 2 * sqrt((I_out)*(I_out) + (Q_out)*(Q_out)) ;
                phase = 57.29 * atan2(Q_out,I_out) ;
                //printf("%5.0f  %5.3f  %5.1f\n\r", Fout, mag, phase) ;
                // log variables to plot
                float log_mag, log_freq;
                log_mag = -gain_mag * log10(mag) + zero_mag ;
                log_freq = decade_width * (log10(Fout)-1) + decade_start ;
                // magnitude bode plot
                drawPixel(log_freq, log_mag, plot_color) ;
                // phase plot
                drawPixel(log_freq, -gain_phase*phase+zero_phase, plot_color) ;

                // the next frequency on log scale
                Fout = Fout * F_inc;

                PT_YIELD_INTERVAL(10) ;
            }
            new_value = false ;
        }
        main_inc = 0;

        PT_YIELD_INTERVAL(20000); 
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

void erase_screen(void){
     fillRect(0, 15, 319, 239, BLACK);  
            // set up the graphics
          //
          drawHLine(10, 120, 300, rgb(1,1,1));
          drawHLine(10, 230, 300, rgb(1,1,1));
          // graph Frequency lines
          for (int j=0; j<4; j++){
            drawVLine(20+j*decade_width, 20, 220, rgb(1,1,1));
            drawVLine(20+j*decade_width+23, 20, 210, rgb(1,1,1));
            drawVLine(20+j*decade_width+52, 20, 210, rgb(1,1,1));
            setTextColor(GRAY2) ;
            setCursor(10+j*decade_width, 231) ;
            setTextSize(1) ;
            sprintf(video_string, "%-5d", decade_name[j]) ;
            writeString(video_string) ;
          }
          // mag lines
          for (int j=0; j<5; j++){
            drawHLine(20, zero_mag+j*gain_mag, 280, rgb(1,1,1));
            setTextColor(rgb(0,5,0)) ;
            setCursor(10, zero_mag+j*gain_mag) ;
            setTextSize(1) ;
            sprintf(video_string, "%-5.4f", mag_name[j]) ;
            writeString(video_string) ;
          }
          // phase lines
          // mag lines
          drawHLine(20, zero_phase, 280, rgb(1,1,1));
          setTextColor(rgb(0,5,2)) ;
          setCursor(10, zero_phase) ;
          setTextSize(1) ;
          writeString("O") ;
          drawHLine(20, zero_phase+90*gain_phase, 280, rgb(1,1,1));
          drawHLine(20, zero_phase+180*gain_phase, 280, rgb(1,1,1));
          setCursor(10, zero_phase+180*gain_phase) ;
          writeString("-18O") ;
          drawHLine(20, zero_phase-90*gain_phase, 280, rgb(1,1,1));
          drawHLine(20, zero_phase-180*gain_phase, 280, rgb(1,1,1));
          setCursor(10, zero_phase-180*gain_phase) ;
          writeString("+18O");
            
    }

// ==================================================
// === user's serial input thread on core 0
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
s1x14 fir_filter(s1x14 x, s1x14 b[], int filter_length);
#define pi 3.14159
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
        static char video_buffer[20];
        static char cmd[16]="erase", arg1[16], arg2[16], arg3[16], arg4[16], arg5[16], arg6[16] ;
        static char* token ;
        static s1x14 one_fix ;
        one_fix = float_to_s1x14(1.0);
      //
      printf("Type 'help' for commands\n\r") ;
      erase_screen();

      while(1) {
        PT_YIELD_usec(100000);
        PT_YIELD_UNTIL(pt, new_value==false) ;

        // print prompt
        sprintf(pt_serial_out_buffer, "filter cmd> ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        //sscanf(pt_serial_in_buffer, "%s %f %f %f %f", arg0, &arg1, arg2, arg3, arg4) ;
        // tokenize
        token = strtok(pt_serial_in_buffer, "  ");
        strcpy(cmd, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg1, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg2, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg3, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg4, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg5, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg6, token) ;


        // parse by command
        if(strcmp(cmd,"help")==0){
            // filter design commands
            printf("***\n\rlp  cutoff_freq length window_type(r,h,b) \n\r"); 
            printf("hp cutoff_freq length window_type\n\r");  
            printf("bp lo_cutoff_freq  hi_cutoff_freq length window_type\n\r");           
            printf("boxcar length \n\r");
            //printf("hp2lp2 hp_cutoff_freq1 Q1 lp_cutoff_freq2 Q2 -- bandpass-SOS 4-pole\n\r***\n\r"); 
            // control commands  
            printf("freq low high ratio_increment -- freqs in Hz, inc>1\n\r");
           // printf("linfreq low high linear_increment -- freqs in Hz, inc>1\n\r");
            printf("color  single_letter -- g, r, b, c, m, w, y\n\r");
            printf("downsample yes/no -- 4x sample slowdown\n\r");
            printf("erase -- clears plot, draws log axes\n\r") ;
           // printf("linear -- draws a linear plot\n\r") ;
            // realtime filtering control
            printf("realtime yes/no -- turn on/off audio filtering\n\r");
            printf("dac yes/no -- turn on/off dac during audio filtering\n\r***\n\r");
        }
        //
        else if(strcmp(cmd,"erase")==0){
            erase_screen(); 
        }
        //
        else if(strcmp(cmd,"realtime")==0){
            if(strcmp(arg1,"yes")==0)  audio_flag = 1;
            else audio_flag = 0;
        }

        //
        else if(strcmp(cmd,"dac")==0){
            if(strcmp(arg1,"yes")==0)  dac_flag = 1;
            else dac_flag = 0;
        }
        //
        else if(strcmp(cmd,"downsample")==0){
            if(strcmp(arg1,"yes")==0)  decimate = 1;
            else decimate = 0;
        }
        //
        else if(strcmp(cmd,"color")==0){
            if(strcmp(arg1,"g")==0)  plot_color = GREEN;
            if(strcmp(arg1,"r")==0)  plot_color = RED;
            if(strcmp(arg1,"b")==0)  plot_color = BLUE;
            if(strcmp(arg1,"c")==0)  plot_color = CYAN;
            if(strcmp(arg1,"m")==0)  plot_color = MAGENTA;
            if(strcmp(arg1,"w")==0)  plot_color = WHITE;
            if(strcmp(arg1,"y")==0)  plot_color = YELLOW;
        }
        //
        else if(strcmp(cmd,"freq")==0){
            sscanf(arg1,"%f", &F_low);
            sscanf(arg2,"%f", &F_high);
            sscanf(arg3,"%f", &F_inc);
            // some error checking
            if(F_low < 10) F_low = 10 ;
            if(F_high > Fs/2) F_high = Fs/2 ;
            if(F_inc <= 1.01) F_inc = 1.1 ;
        }
        // 
        
        else if(strcmp(cmd,"boxcar")==0){
            sscanf(arg1,"%d", &filter_length);
            // some error checking
            if(filter_length < 3) filter_length = 3 ;
            if(filter_length > 120) filter_length = 120;
            for(int i=0; i<filter_length; i++ ){
              b[i] = float_to_s1x14(1.0 / (float)filter_length) ;
            }
            // signal the plotting thread
            filter_type = 1;
            new_value = true ;
        }  
        // 
        else if(strcmp(cmd,"lp")==0){
            sscanf(arg1,"%f", &F0);
            sscanf(arg2,"%d", &filter_length);
            sscanf(arg3,"%c", &window_type);

            // some error checking
            if(filter_length < 3) filter_length = 3 ;
            if(filter_length > 120) filter_length = 120;
            // odd length filters only!
            if((filter_length & 1) == 0) filter_length += 1;
          //printf("%d\n\r", filter_length) ;

            w0 = 2*pi*F0/Fs ; //Ft
            M = filter_length - 1 ;
            for(int i=0; i<filter_length; i++){
              if(i == (int)M/2) b_float[i] = (2*F0/Fs);
              else b_float[i] = (sin(w0 * (i-M/2)) / (pi*(i-M/2))) ;

              if(window_type=='r') wind_val = (1.0); // / (float)filter_length
              if(window_type=='h') wind_val = (0.54 - 0.46 * cos(2*pi*i/M));
              if(window_type=='b') wind_val = (0.42 - 0.5 * cos(2*pi*i/M) + 0.08 * cos(4*pi*i/M));

              if(i <= M) b_float[i] = b_float[i] * wind_val ;
              else b_float[i] = 0 ;

              b[i] = float_to_s1x14(b_float[i]) ;
              //printf("%d %f\n\r", i, s1x14_to_float(b[i])) ;  
            }
            
            // signal the plotting thread
            filter_type = 1;
            new_value = true ;
        }  
        //
        else if(strcmp(cmd,"hp")==0){
            sscanf(arg1,"%f", &F0);
            sscanf(arg2,"%d", &filter_length);
            sscanf(arg3,"%c", &window_type);
            // some error checking
            if(filter_length < 3) filter_length = 3 ;
            if(filter_length > 120) filter_length = 120;
            // odd length filters only!
            if((filter_length & 1) == 0) filter_length += 1;
          //printf("%d\n\r", filter_length) ;

            w0 = 2*pi*F0/Fs ; //Ft
            M = filter_length - 1 ;
            for(int i=0; i<filter_length; i++){
              if(i == (int)M/2) b_float[i] = 1 - (2*F0/Fs);
              else b_float[i] = - (sin(w0 * (i-M/2)) / (pi*(i-M/2))) ;

              if(window_type=='r') wind_val = (1.0); // / (float)filter_length
              if(window_type=='h') wind_val = (0.54 - 0.46 * cos(2*pi*i/M));
              if(window_type=='b') wind_val = (0.42 - 0.5 * cos(2*pi*i/M) + 0.08 * cos(4*pi*i/M));

              if(i <= M) b_float[i] = b_float[i] * wind_val ;
              else b_float[i] = 0 ;

              b[i] = float_to_s1x14(b_float[i]) ;
              //printf("%d %f\n\r", i, s1x14_to_float(b[i])) ;  
            }
            
            // signal the plotting thread
            filter_type = 1;
            new_value = true ;
        }  
        
        //
        else if(strcmp(cmd,"bp")==0){
            sscanf(arg1,"%f", &F1);
            sscanf(arg2,"%f", &F2);
            sscanf(arg3,"%d", &filter_length);
            sscanf(arg4,"%c", &window_type);
            // some error checking
            if(filter_length < 3) filter_length = 3 ;
            if(filter_length > 120) filter_length = 120;
            // odd length filters only!
            if((filter_length & 1) == 0) filter_length += 1;
          //printf("%d\n\r", filter_length) ;

            w1 = 2*pi*F1/Fs ; //Ft1
            w2 = 2*pi*F2/Fs ; //Ft2
            M = filter_length - 1 ;
            for(int i=0; i<filter_length; i++){
              if(i == (int)M/2) b_float[i] = (2*F2/Fs) - (2*F1/Fs);
              else b_float[i] =  (sin(w2 * (i-M/2)) / (pi*(i-M/2))) - (sin(w1 * (i-M/2)) / (pi*(i-M/2))) ;

              if(window_type=='r') wind_val = (1.0); // / (float)filter_length
              if(window_type=='h') wind_val = (0.54 - 0.46 * cos(2*pi*i/M));
              if(window_type=='b') wind_val = (0.42 - 0.5 * cos(2*pi*i/M) + 0.08 * cos(4*pi*i/M));

              if(i <= M) b_float[i] = b_float[i] * wind_val ;
              else b_float[i] = 0 ;

              b[i] = float_to_s1x14(b_float[i]) ;
              //printf("%d %f\n\r", i, s1x14_to_float(b[i])) ;  
            }
            
            // signal the plotting thread
            filter_type = 1;
            new_value = true ;
        }  

        //
        
        else printf("Huh?\n\r") ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ==========================================
// === set up timer ISR  used in this pgm
// ==========================================
// === timer alarm ========================
// !! modifiying alarm zero trashes the cpu 
//        and causes LED  4 long - 4 short
// !! DO NOT USE alarm 0
// This low-level setup is ocnsiderably faster to execute
// than the hogh-level callback

#define ALARM_NUM 1
#define ALARM_IRQ TIMER_IRQ_1
// ISR interval will be 10 uSec
//
// the actual ISR
void compute_sample(void);
//
static void alarm_irq(void) {
    // mark ISR entry
    gpio_put(2,1);
    // Clear the alarm irq
    hw_clear_bits(&timer_hw->intr, 1u << ALARM_NUM);
    // arm the next interrupt
    // Write the lower 32 bits of the target time to the alarm to arm it
    timer_hw->alarm[ALARM_NUM] = timer_hw->timerawl + alarm_period ;
    //
    compute_sample();

    // mark ISR exit
    gpio_put(2,0);
}

// set up the timer alarm ISR
static void alarm_in_us(uint32_t delay_us) {
    // Enable the interrupt for our alarm (the timer outputs 4 alarm irqs)
    hw_set_bits(&timer_hw->inte, 1u << ALARM_NUM);
    // Set irq handler for alarm irq
    irq_set_exclusive_handler(ALARM_IRQ, alarm_irq);
    // Enable the alarm irq
    irq_set_enabled(ALARM_IRQ, true);
    // Enable interrupt in block and at processor
    // Alarm is only 32 bits 
    uint64_t target = timer_hw->timerawl + delay_us;
    // Write the lower 32 bits of the target time to the alarm which
    // will arm it
    timer_hw->alarm[ALARM_NUM] = (uint32_t) target;   
}

//======================================================== 
// FIR filter
// assumes ODD order, symmetric b[]
// about 60 taps max at 40 KHz
// =======================================================
s1x14 fir_filter(s1x14 x, s1x14 b[], int filter_length){
  static s1x14 y;
  static s1x14 x_buf[128];
  static int i, ptrin=1, ptrout=2 ;

  // update input history
  memmove(x_buf+1, x_buf, 2*(filter_length-1));
  x_buf[0] = x;

  y = 0 ;
  // assumes filter_length is ODD
  int fl2 = filter_length/2 ;
  for(i=0; i<(fl2); i++){
    y = y + muls1x14(b[i], x_buf[i] + x_buf[filter_length-i-1]) ;
  }
  //gpio_put(2,0);
  y = y + muls1x14(b[fl2], x_buf[fl2]) ;
  //gpio_put(2,1);
  return y;
}

//======================================================== 
// *************** do not change *****************
// Second order cheby1 lowpass
// NOTE: this is the compensator for the downsampler and
// should NOT be changed!
// =======================================================
// xx is the current input signal sample
// returns the current filtered output sample
// Simplification: b(1)*x(n)+b(2)*x(n-1)+b(3)*x(n-2) = 
// b(1)*x(n)+2*b(1)*x(n-1)+b(1)*x(n-2) =
// b(1)* (x(n)+(x(n-1)<<1)+x(n-2))
//  DO NOT CHANGE
  s1x14 CIC_ChebyOne2low(s1x14 xx){
  static s1x14 yy  ; 
  //IIR state variables
  static s1x14  b1,  a2, a3 ;
  static s1x14  xn, xn_1, xn_2 ; 
  static s1x14  yn_1, yn_2  ;

  // CIC compensator
  // [b,a] = cheby1(2, 8, 0.5)
  b1 = 1927  ;            
  a2 = -8079  ;  
  a3 = 11053 ;  

  // sum the 5 terms: yy += xx*coeff 
  // and update the state variables
  // as soon as possible

  yy = muls1x14((xx+(xn_1<<1)+xn_2), b1);
  xn_2 = xn_1;
  xn_1 = xx;
  yy = yy - muls1x14(yn_2, a3); 
  yn_2 = yn_1;
  yy = yy - muls1x14(yn_1, a2); 
  yn_1 = yy;
  return yy;
} 

//========================================================   
// 4th order CIC
// / CIC filter
// N=4, R=4, M=2   
// -- N=order=4 (4 integrators)
// -- R=downsample=4 -- M=delay(after downsample)=2
// output rate is 10 KHz sample rate

s1x14 CIC_n4_r4_m2(s1x14 xx)
{
     //CIC state variables
 static int32_t integrator1, integrator2, integrator3, integrator4 ;
 static int32_t  last_integrator4, last2_integrator4 ;
 static int32_t  last_integrator2, last2_integrator2 ;
 static int32_t  comb1, comb2, comb3, comb4 ;
 static int32_t  last_comb1, last_comb2, last_comb3, last_comb4 ; 
 static int32_t  last2_comb1, last2_comb2, last2_comb3, last2_comb4 ; 
 //https://dsp.stackexchange.com/questions/19584/how-to-make-cic-compensation-filter
 static s1x14 yy, zz ;

     integrator1 = integrator1 + xx ;
	 integrator2 = integrator2 + integrator1 ;
	 integrator3 = integrator3 + integrator2 ;
	 integrator4 = integrator4 + integrator3 ;
	 
	 if ((downsample_clock & 0x03) == 0) //modulo 4
	 {
		comb1 = integrator4 - last2_integrator4 ;
    	comb2 = comb1 - last2_comb1 ;
    	comb3 = comb2 - last2_comb2 ;
        //
    	yy = (s1x14)((comb3 - last2_comb3)>>11) ; //16
		
		// update state
    	last2_integrator4 = last_integrator4 ;
    	last_integrator4 = integrator4 ;
    	last2_comb1 = last_comb1 ;
    	last_comb1 = comb1;
    	last2_comb2 = last_comb2 ;
    	last_comb2 = comb2;
    	last2_comb3 = last_comb3 ;
    	last_comb3 = comb3;
        // this is a compensator to flatten downsample spectrum
        // increases bandwidth from 0.2 to 0.4 of NEW Nyquist freq
        // [b,a] = cheby1(2, 8, 0.5)
        zz = CIC_ChebyOne2low(yy) ;
	 }
    return zz;
} 

// ==================================================
// === dsp ISR -- RUNNING on core 1 -- 40 KHz
// Nyquist frequency is 20 KHz
// ==================================================
// 
void compute_sample(void)
{
    // make the sine wave to send to filter
    main_accum += main_inc ;
    // update main waveform
    sine_table_out = sine_table[main_accum>>24] ;
    cos_table_out = cos_table[main_accum>>24] ;

    // read the ADC chnnel 0
    adc_in = adc_read();
    // choose the actual input to filters
    // and convert adc input to fixed point normalized to 12 bits
    if(audio_flag) filter_input = adc_in <<1 ; //audio_flag
    else filter_input = sine_table_out ;

    // Now the filter
    // with  timing by gpio pin
    //gpio_put(2,0);
    if(decimate){
        // down sampled signal updated every 4th ISR
        down_sampled_signal =  CIC_n4_r4_m2(filter_input) ;
        // all four phases of the downsampler can be used for different
        // filters: phase 0 is used by the downsample process itself
        if((downsample_clock & 0x03) == 1){
            switch(filter_type){
                case 0: {filter_out =(down_sampled_signal); break;}
                case 1: {filter_out = fir_filter(down_sampled_signal, b, filter_length); break;}
            }
        }
    }
    else {
        switch(filter_type){
            case 0: {filter_out =(filter_input); break;}
            case 1: {filter_out = fir_filter(filter_input, b, filter_length); break;}
        }
    }
    //gpio_put(2,1);

    if (audio_flag | dac_flag){
      DAC_data = (DAC_config_chan_A | (((filter_out>>3) + 2048) & 0xfff))  ; //14
      spi_write16_blocking(SPI_PORT, &DAC_data, 1) ;
      DAC_data = (DAC_config_chan_B | (((filter_input>>3) + 2048) & 0xfff))  ;
      spi_write16_blocking(SPI_PORT, &DAC_data, 1) ;
    }  

    if(audio_flag==0) {
      // do quadrature multiplies for amp and phase
      I_signal = muls1x14(filter_out, sine_table_out );
      Q_signal = muls1x14(filter_out, cos_table_out );
      // accumulate signal for averaging
      // main thread waits for count to reach maximum value
      if (avg_count<avg_size) {
          I_avg += I_signal ;
          Q_avg += Q_signal ;
          avg_count++ ;
      }
    }

    // modulo counter for downsampling
    // by 4, so new Nyquist is 5 KHz for decimaed filters
    downsample_clock++ ;

} // isr sample routine

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  // testing timing
    gpio_init(2) ;	
    gpio_set_dir(2, GPIO_OUT) ;

    // init the adc
  adc_init();
  // Make sure GPIO is high-impedance, no pullups etc
  adc_gpio_init(26);
  // Select ADC input 0 (GPIO26)
  adc_select_input(0);

  // fire off interrupt
    alarm_in_us(alarm_period);
    while(1) {};
  //
  // === initalize the scheduler ==========
  //pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;

  // Initialize SPI channel (channel, baud rate set to 20MHz)
    // connected to spi DAC
    spi_init(SPI_PORT, 20000000) ;
    // Format (channel, data bits per transfer, polarity, phase, order)
    spi_set_format(SPI_PORT, 16, 0, 0, 0);
    // Map SPI signals to GPIO ports
    //gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_MOSI, GPIO_FUNC_SPI);
    gpio_set_function(PIN_CS, GPIO_FUNC_SPI) ;

    int i;
    for (i=0; i<256; i++) {
      // sine table is in natural +1/-1 range
      sine_table[i] = float_to_s1x14(0.95 * sin(2*3.1416*(float)i/256)) ;
      cos_table[i] = float_to_s1x14(0.95 * cos(2*3.1416*(float)i/256)) ;
    }
   // printf("%f\n\r", s15x16_to_float(sine_table[64])) ;

  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main